
struct Context
{
	Context*      parent;
	string  var_extended;
	InXML* node_extended;
	set<InXML*>*  posbnd;
	InXML    *root, *ctx;

	Context( InXML* root ) : parent(0), root(root), ctx(root), posbnd(0) {}
	Context( Context* p, const string& v, InXML* n, set<InXML*>* pb )
		: parent(p), var_extended(v), node_extended(n), root(p->root), ctx(n), posbnd(pb) {}
	InXML* operator[]( const string& var ) {
		return var==var_extended ? node_extended : (*parent)[var];
	}
	int never_bind( const string& var, InXML* ix ) {
		return var==var_extended ? !posbnd->count(ix) : parent?parent->never_bind(var,ix):0;
	}
};


struct QueryCache
{
	map< Automaton*, set<InXML*> >              possibly_bound;
	map< Automaton*, vector< vector<InXML*> > > query_result;
};


struct SingleQueryResult : mem_man<SingleQueryResult>
{
	enum Type {Empty, All, Range};
	Type t;
};



struct Engine
{
public:
	Engine( const char* pg_dir ) : main_template(parse_template(pg_dir)) {}
	vector<OutXML*> execute( InXML* root );

private:
	vector<Node> main_template;
	QueryCache   qc;

public:
	void eval( vector<OutXML*>& result, vector<Node>& ns, Context& ctx )
	{
		for(int i=0; i!=ns.size(); ++i)
			eval( result, ns[i], ctx );
	}
	void eval( vector<OutXML*>& result, Node& n, Context& ctx )
	{
		if( as<Elem>(n) )    return eval( result, *as<Elem>(n), ctx );
		if( as<Attr>(n) )    return eval( result, *as<Attr>(n), ctx );
		if( as<DynElem>(n) ) return eval( result, *as<DynElem>(n), ctx );
		if( as<DynAttr>(n) ) return eval( result, *as<DynAttr>(n), ctx );
		if( as<Pcdata>(n) )  return eval( result, *as<Pcdata>(n), ctx );
		if( as<Var>(n) )     return eval( result, *as<Var>(n), ctx );
		if( as<Gather>(n) )  return eval( result, *as<Gather>(n), ctx );
		if( as<Visit>(n) )   return eval( result, *as<Visit>(n), ctx );
	}
	void eval( vector<OutXML*>& result, Elem& n, Context& ctx )
	{
		vector<OutXML*> C; eval(C, n.C, ctx);
		result.push_back( new OutXML_Element(n.D, C) );
	}
	void eval( vector<OutXML*>& result, Attr& n, Context& ctx )
	{
		vector<OutXML*> C; eval(C, n.C, ctx);
		result.push_back( new OutXML_Attribute(n.D, C) );
	}
	void eval( vector<OutXML*>& result, DynElem& n, Context& ctx )
	{
		vector<OutXML*> C; eval(C, n.C, ctx);
		result.push_back( new OutXML_Element(ctx[n.D]->D, C) );
	}
	void eval( vector<OutXML*>& result, DynAttr& n, Context& ctx )
	{
		vector<OutXML*> C; eval(C, n.C, ctx);
		result.push_back( new OutXML_Attribute(ctx[n.D]->D, C) );
	}
	void eval( vector<OutXML*>& result, Pcdata& n, Context& ctx )
	{
		result.push_back( new OutXML_Text(n.D) );
	}
	void eval( vector<OutXML*>& result, Var& n, Context& ctx )
	{
		result.push_back( ctx[n.D]->as_OutXML() );
	}
	void eval( vector<OutXML*>& result, Gather& n, Context& ctx )
	{
		set<InXML*> visited;
		for(int i=0; i!=n.U.size(); ++i)
		{
			vector<InXML*> qr;
			if( !query_list( qr, n.V, n.U[i]->A, ctx, qc ) )
				set_all_nodes(qr, ctx.root);

			for(int j=0; j!=qr.size(); ++j)
				if( !visited.count(qr[j]) ) {
					visited.insert(qr[j]);
					Context c2(&ctx, n.V, qr[j], &qc.possibly_bound[&n.U[i]->A]);
					eval( result, n.U[i]->C, c2 );
				}
		}
	}
	void eval( vector<OutXML*>& result, Visit& n, Context& ctx )
	{
		struct visit_rec
		{
			Engine*                       pThis;
			vector<Node>*          default_tmpl;
			map<InXML*, vector<Node>*> tmpl_for;
			Context*                        ctx;
			set<InXML*>                 visited;
			string                            v;
			map<InXML*, set<InXML*>*>    posbnd;

			vector<Node>* should_visit( InXML* ix ) {
				if( !ix || visited.count(ix)  )              return 0;
				//if( !ctx->ctx->is_anscestor_or_self_of(ix) ) return 0;
				return tmpl_for.count(ix) ? tmpl_for[ix] : default_tmpl;
			}

			bool f( vector<OutXML*>& r, OutXML* ox )
			{
				InXML* ix = ox->as_InXML();
				if( vector<Node>* tmpl = should_visit(ix) ) {
					visited.insert(ix);
						Context c2(ctx, v, ix, posbnd[ix]);
						vector<OutXML*> r1;
						pThis->eval( r1, *tmpl, c2 );
						for(int i=0;i!=r1.size();++i)
							f( r, r1[i] );
					visited.erase(ix);
					return true;
				}
				vector<OutXML*> r1; bool changed=false;
				for(int i=0;i!=ox->C.size();++i)
					changed |= f( r1, ox->C[i] );
				r.push_back( changed ? ox->clone_with(r1) : ox );
				return changed;
			}
		}v;

		v.pThis=this;
		v.default_tmpl = 0;
		v.v   = n.V;
		v.ctx = &ctx;
		for(int i=0; i!=n.U.size(); ++i)
		{
			vector<InXML*> qr;
			if( !query_list( qr, n.V, n.U[i]->A, ctx, qc ) )
				{v.default_tmpl = &n.U[i]->C; break;}
			for(int j=0; j!=qr.size(); ++j)
				if( !v.tmpl_for.count(qr[j]) ) {
					v.posbnd[qr[j]] = &qc.possibly_bound[&n.U[i]->A];
					v.tmpl_for[qr[j]] = &n.U[i]->C;
				}
		}

		// v.f( result, ctx.ctx->as_OutXML() );
		if( n.R == "-" )
			v.f( result, ctx.root->as_OutXML() );
		else
			v.f( result, ctx[n.R]->as_OutXML() );
	}

private:
	bool query_list( vector<InXML*>& qr, const string& v,
                 Automaton& a, Context& ctx, QueryCache& qc );
	void set_all_nodes( vector<InXML*>& qr, InXML* root );
};
