extern "C" {
//#include <Lib/gtalib.h>
#include "mona/gtalib.h"
}

struct Automaton
{
	mgGta*                 gta;
	map<string,int>     varMap; // (ラベルも含む)変数名からindexへのマップ
	vector<string> realVarName; // (ラベルを含まない)変数名リスト
	vector<int>     realVarMap; // (ラベルを含まない)変数idxから(ラベルも含む)idx
	vector<int>  invRealVarMap; // (ラベルを含む)変数idxから(ラベルを含まない)idx
	set<mgState>        finals;
	vector< pair<int,int> > pairedVars; // イコールでないといけない変数idxペア

	Automaton( const char* gtaFileName )
	{
		gta = mgLoad( const_cast<char*>(gtaFileName) );
		for( int i=0; i!=gta->numVars; ++i )
		{
			string s(gta->var[i].name);
			varMap[s] = i;
			if( s.size()<3 || s[0]!='_' || s[1]!='_' || s[2]=='q' ) {
				invRealVarMap.push_back(realVarMap.size());
				realVarName.push_back(s);
				realVarMap.push_back(i);
			} else {
				invRealVarMap.push_back(-1);
			}
		}

		vector<char> zeros(gta->numVars);
		for(mgState s=0; s!=gta->stateSpace[1].numStates; ++s)
			if( gta->final[mgDelta(gta,0,s,gta->stateSpace[0].initial,&zeros[0])]
			  == mgACCEPT )
				finals.insert(s);

		for(int i=0; i!=realVarName.size(); ++i)
			if( realVarName[i].substr(0,3) == "__q"
			 && realVarName[i][realVarName[i].size()-1]=='a' ) {
				string rvb = realVarName[i];
				rvb[rvb.size()-1] = 'b';
				pairedVars.push_back( make_pair(varMap[realVarName[i]],varMap[rvb]) );
			}

		calc_mask_of_state();
	}

	~Automaton()
	{
		if( gta )
			mgFree(gta);
	}

	int varNum()
	{
		return gta->numVars;
	}

	mgState init()
	{
		return gta->stateSpace[1].initial;
	}

	mgState delta( mgState L, mgState R, char* A )
	{
		return mgDelta( gta, 1, L, R, A );
	}

	bool is_final(mgState s)
	{
		return finals.count(s)!=0;
	}

	bool is_tautology()
	{
		return gta->stateSpace[1].numStates==1 && is_final(init());
	}

	int size()
	{
		return gta->stateSpace[1].numStates;
	}

	//
	// q |--> {0,1}^* must have surjective correspondance
	// 
	typedef unsigned int SET_T;
	map<mgState, SET_T> mask_of_state;
	void calc_mask_of_state()
	{
		mgState dame = 0xffffffff;
		mask_of_state[init()] = 0;
		bool changed;
		do {
			changed = false;
			for(map<mgState, SET_T>::iterator it=mask_of_state.begin(); it!=mask_of_state.end(); ++it)
			for(map<mgState, SET_T>::iterator jt=mask_of_state.begin(); jt!=mask_of_state.end(); ++jt)
			for(SET_T s=0; s<(1<<realVarName.size()); ++s)
			if( it->first != dame && jt->first != dame )
			if( !(it->second & jt->second) && !(s & (it->second|jt->second)) )
			{
				set<mgState> q = abstract_delta( it->first, jt->first, s );
				for(set<mgState>::iterator kt=q.begin(); kt!=q.end(); ++kt)
				if( *kt != dame )
				{
					SET_T ss = s | it->second | jt->second;
					if( !mask_of_state.count(*kt) )
						changed=true, mask_of_state[*kt] = ss;
					else if( mask_of_state[*kt] != ss )
						{assert(dame==-1); changed=true, dame = *kt;}
				}
			}
		} while( changed );
		mask_of_state.erase(dame);
	}

	set<mgState> abstract_delta( mgState L, mgState R, SET_T s )
	{
		mBdd* bdd = gta->stateSpace[1].bddNode;
		mNode   n = gta->stateSpace[1].behaviour[L][R];
		set<mNode> visited;
		set<mgState> reachable;
		abstract_delta_rec(bdd, n, s, visited, reachable);
		return reachable;
	}

	void abstract_delta_rec( mBdd* bdd, mNode n, SET_T s, set<mNode>& visited, set<mgState>& reachable )
	{
		if( visited.count(n) )
			return;

		if( bdd[n].idx == -1 )
			{ reachable.insert( bdd[n].lo ); return; }

		visited.insert( n );
		int i = invRealVarMap[bdd[n].idx];
		if( i == -1 )
			abstract_delta_rec( bdd, bdd[n].lo, s, visited, reachable ),
			abstract_delta_rec( bdd, bdd[n].hi, s, visited, reachable );
		else
			abstract_delta_rec( bdd, s&(1<<i) ? bdd[n].hi : bdd[n].lo, s, visited, reachable );
	}
};
