#include "common.h"
#include <cctype>
#include <iterator>
#include <fstream>

struct Lexer
{
	const char* p;
	Lexer( const char* tmpl ) : p(tmpl) {}

	void skip_spaces() {
		while( isspace(*p) )
			++p;
	}
	char eat_cmd() {
		skip_spaces();
		return isalpha(*p) ? *p++ : -1;
	}
	char eat_char() {
		skip_spaces();
		return *p ? *p++ : -1;
	}
	int eat_int() {
		skip_spaces();
		int i=0;
		while( isdigit(p[i]) )
			++i;
		int n = i ? atoi(string(p,p+i).c_str()) : -1;
		p += i;
		return n;
	}
	string eat_name() {
		skip_spaces();
		int i=0;
		while( p[i] && !isspace(p[i]) )
			++i;
		string s(p, p+i);
		p += i;
		return s;
	}
	string eat_string() {
		skip_spaces();
		string ans;
		for(++p;;)
			switch(*p) {
			case '\\':
				switch(p[1]) {
				case 't':  ans+='\t'; break;
				case 'r':  ans+='\r'; break;
				case 'n':  ans+='\n'; break;
				default:   ans+=p[1]; break;
				} p+=2; break;
			case '&':  ans+="&amp;", ++p; break;
			case '<':  ans+="&lt;",  ++p; break;
			case '>':  ans+="&gt;",  ++p; break;
			case '"':  ++p; return ans;
			default:   ans+=*p++; break;
			}
		return ans;
	}
};

struct Parser
{
	string       dir;
	vector<Node> nod;

	Parser( const char* pg_dir )
		: dir(pg_dir+string("/"))
	{
		ifstream fin( (dir+"template").c_str() );
		fin.unsetf(ios::skipws);
		istream_iterator<char> inp(fin), ut;
		vector<char> tmpl(inp, ut);
		tmpl.push_back(0);

		Lexer lex(&tmpl[0]);
		parse_list( lex, nod );
	}

	void parse_list( Lexer& lex, vector<Node>& result )
	{
		for(;;)
			switch( lex.eat_cmd() )
			{
			case 'g':
				{ string v = lex.eat_name(); //ほんとは複数来うる
				  lex.eat_char();            // "{"
				  vector<GVUnit> gv;
				  for(;;) {
				    int aid = lex.eat_int();
				    if( aid<0 ) break;
				    char a[256]; sprintf(a, "%s%d", dir.c_str(), aid);
				    vector<Node> c;
				    parse_list(lex,c);
				    gv.push_back( GVUnit(new GVUnit_t(a,c)) );
				  }
				  result.push_back(Node(new Gather(v,gv)));
				  lex.eat_char();  // "}"
				  break; }
			case 'v':
				{ string v = lex.eat_name(); //ほんとは複数来うる
				  string r = lex.eat_name();
				  lex.eat_char();            // "{"
				  vector<GVUnit> gv;
				  for(;;) {
				    int aid = lex.eat_int();
				    if( aid<0 ) break;
				    char a[256]; sprintf(a, "%s%d", dir.c_str(), aid);
				    vector<Node> c;
				    parse_list(lex,c);
				    gv.push_back( GVUnit(new GVUnit_t(a,c)) );
				  }
				  result.push_back(Node(new Visit(v,r,gv)));
				  lex.eat_char();  // "}"
				  break; }
			case 'e':
				{ string name = lex.eat_name();
				  lex.eat_char();
				  vector<Node> c; parse_list(lex,c);
				  result.push_back(Node(new Elem(name,c)));
				  lex.eat_char();
				  break; }
			case 'a':
				{ string name = lex.eat_name();
				  lex.eat_char();
				  vector<Node> c; parse_list(lex,c);
				  result.push_back(Node(new Attr(name,c)));
				  lex.eat_char();
				  break; }
			case 's':
				{ string str = lex.eat_string();
				  result.push_back(Node(new Pcdata(str)));
				  break; }
			case 'E':
				{ string name = lex.eat_name();
				  lex.eat_char();
				  vector<Node> c; parse_list(lex,c);
				  result.push_back(Node(new DynElem(name,c)));
				  lex.eat_char();
				  break; }
			case 'A':
				{ string name = lex.eat_name();
				  lex.eat_char();
				  vector<Node> c; parse_list(lex,c);
				  result.push_back(Node(new DynAttr(name,c)));
				  lex.eat_char();
				  break; }
			case 'X':
				{ string name = lex.eat_name();
				  result.push_back(Node(new Var(name)));
				  break; }
			default:
				return;
			}
	}
};

vector<Node> parse_template( const char* dir )
{
	return Parser(dir).nod;
}
